{% set siblings = this.get_siblings() %}
{% if siblings.prev_page %}<a href="{{ siblings.prev_page|url }}">previous</a>{% endif %}
{% if siblings.next_page %}<a href="{{ siblings.next_page|url }}">next</a>{% endif %}
